from glue.plugins.wcs_autolinking import ASTROPY_GE_31

from glue.config import autolinker, link_helper
from glue.core.link_helpers import MultiLink
from glue.core.coordinates import WCSCoordinates
from glue.utils import efficient_pixel_to_pixel

__all__ = ['IncompatibleWCS', 'WCSLink', 'wcs_autolink']


class IncompatibleWCS(Exception):
    pass


def get_cids_and_functions(wcs1, wcs2, pixel_cids1, pixel_cids2):

    def forwards(*pixel_input):
        return efficient_pixel_to_pixel(wcs1, wcs2, *pixel_input)

    def backwards(*pixel_input):
        return efficient_pixel_to_pixel(wcs2, wcs1, *pixel_input)

    pixel_input = (0,) * len(pixel_cids1)

    try:
        forwards(*pixel_input)
        backwards(*pixel_input)
    except Exception:
        return None, None, None, None

    return pixel_cids1, pixel_cids2, forwards, backwards


@link_helper(category='Astronomy')
class WCSLink(MultiLink):
    """
    A collection of links that link the pixel components of two datasets via
    WCS transformations.
    """

    display = 'WCS link'
    cid_independent = True

    def __init__(self, data1=None, data2=None, cids1=None, cids2=None):

        # Extract WCS objects - from here onwards, we assume that these objects
        # have the new Astropy APE 14 interface.
        wcs1, wcs2 = data1.coords.wcs, data2.coords.wcs

        forwards = backwards = None
        if wcs1.pixel_n_dim == wcs2.pixel_n_dim and wcs1.world_n_dim == wcs2.world_n_dim:
            if (wcs1.world_axis_physical_types.count(None) == 0 and
                    wcs2.world_axis_physical_types.count(None) == 0):

                # The easiest way to check if the WCSes are compatible is to simply try and
                # see if values can be transformed for a single pixel. In future we might
                # find that this requires optimization performance-wise, but for now let's
                # not do premature optimization.

                pixel_cids1, pixel_cids2, forwards, backwards = get_cids_and_functions(wcs1, wcs2,
                                                                                       data1.pixel_component_ids[::-1],
                                                                                       data2.pixel_component_ids[::-1])

                self._physical_types_1 = wcs1.world_axis_physical_types
                self._physical_types_2 = wcs2.world_axis_physical_types

        if not forwards or not backwards:

            # Try setting only a celestial link. We try and extract the celestial
            # WCS, which will only work if the celestial coordinates are separable.
            # TODO: find a more generalized APE 14-compatible way to do this.

            if not wcs1.has_celestial or not wcs2.has_celestial:
                raise IncompatibleWCS("Can't create WCS link between {0} and {1}".format(data1.label, data2.label))

            try:
                wcs1_celestial = wcs1.celestial
                wcs2_celestial = wcs2.celestial
            except Exception:
                raise IncompatibleWCS("Can't create WCS link between {0} and {1}".format(data1.label, data2.label))

            cids1 = data1.pixel_component_ids
            cids1_celestial = [cids1[wcs1.wcs.naxis - wcs1.wcs.lng - 1],
                               cids1[wcs1.wcs.naxis - wcs1.wcs.lat - 1]]

            if wcs1_celestial.wcs.lng > wcs1_celestial.wcs.lat:
                cids1_celestial = cids1_celestial[::-1]

            cids2 = data2.pixel_component_ids
            cids2_celestial = [cids2[wcs2.wcs.naxis - wcs2.wcs.lng - 1],
                               cids2[wcs2.wcs.naxis - wcs2.wcs.lat - 1]]

            if wcs2_celestial.wcs.lng > wcs2_celestial.wcs.lat:
                cids2_celestial = cids2_celestial[::-1]

            pixel_cids1, pixel_cids2, forwards, backwards = get_cids_and_functions(wcs1_celestial, wcs2_celestial,
                                                                                   cids1_celestial, cids2_celestial)

            self._physical_types_1 = wcs1_celestial.world_axis_physical_types
            self._physical_types_2 = wcs2_celestial.world_axis_physical_types

        if pixel_cids1 is None:
            raise IncompatibleWCS("Can't create WCS link between {0} and {1}".format(data1.label, data2.label))

        super(WCSLink, self).__init__(pixel_cids1, pixel_cids2,
                                      forwards=forwards, backwards=backwards)

        self.data1 = data1
        self.data2 = data2

    def __gluestate__(self, context):
        state = {}
        state['data1'] = context.id(self.data1)
        state['data2'] = context.id(self.data2)
        return state

    @classmethod
    def __setgluestate__(cls, rec, context):
        if not ASTROPY_GE_31:
            raise ValueError("Loading this session file requires Astropy 3.1 "
                             "or later to be installed")
        self = cls(context.object(rec['data1']),
                   context.object(rec['data2']))
        return self

    @property
    def description(self):
        types1 = ''.join(['<li>' + phys_type for phys_type in self._physical_types_1])
        types2 = ''.join(['<li>' + phys_type for phys_type in self._physical_types_2])
        return ('This automatically links the coordinates of the '
                'two datasets using the World Coordinate System (WCS) '
                'coordinates defined in the files.<br><br>The physical types '
                'of the coordinates linked in the first dataset are: '
                '<ul>{0}</ul>and in the second dataset:<ul>{1}</ul>'
                .format(types1, types2))


@autolinker('Astronomy WCS')
def wcs_autolink(data_collection):

    # Find subset of datasets with WCS coordinates
    wcs_datasets = [data for data in data_collection
                    if hasattr(data, 'coords') and isinstance(data.coords, WCSCoordinates)]

    # Only continue if there are at least two such datasets
    if len(wcs_datasets) < 2:
        return []

    # Find existing WCS links
    existing = set()
    for link in data_collection.external_links:
        if isinstance(link, WCSLink):
            existing.add((link.data1, link.data2))

    # Loop through all pairs of datasets, skipping pairs for which a link
    # already exists. PERF: in practice we don't actually have to link all
    # pairs, so we should try and optimize that.
    all_links = []
    for i1, data1 in enumerate(wcs_datasets):
        for data2 in wcs_datasets[i1 + 1:]:
            if (data1, data2) not in existing:
                try:
                    link = WCSLink(data1, data2)
                except IncompatibleWCS:
                    continue
                all_links.append(link)

    return all_links



from __future__ import absolute_import, division, print_function

import os
import weakref
import traceback
from collections import OrderedDict

import numpy as np

from qtpy.QtCore import Qt
from qtpy import QtWidgets, QtGui

from glue.utils import color2hex, nanmean, nanmedian, nanmin, nanmax, nansum
from glue.config import fit_plugin, viewer_tool
from glue.utils.qt import load_ui, fix_tab_widget_fontsize
from glue.viewers.profile.qt.mouse_mode import NavigateMouseMode, RangeMouseMode
from glue.core.qt.fitters import FitSettingsWidget
from glue.utils.qt import Worker
from glue.viewers.common.tool import Tool
from glue.viewers.image.state import AggregateSlice
from glue.core.aggregate import mom1, mom2
from glue.core import BaseData, Subset
from glue.viewers.image.qt import ImageViewer
from glue.core.link_manager import is_convertible_to_single_pixel_cid
from glue.external.echo import SelectionCallbackProperty
from glue.external.echo.qt import connect_combo_selection

__all__ = ['ProfileTools']


MODES = ['navigate', 'fit', 'collapse']

COLLAPSE_FUNCS = OrderedDict([(nanmean, 'Mean'),
                              (nanmedian, 'Median'),
                              (nanmin, 'Minimum'),
                              (nanmax, 'Maximum'),
                              (nansum, 'Sum'),
                              (mom1, 'Moment 1'),
                              (mom2, 'Moment 2')])


@viewer_tool
class ProfileAnalysisTool(Tool):

    action_text = 'Options'
    tool_id = 'profile-analysis'

    def __init__(self, viewer):
        super(ProfileAnalysisTool, self).__init__(viewer)
        self._profile_tools = ProfileTools(viewer)
        container_widget = QtWidgets.QSplitter(Qt.Horizontal)
        plot_widget = viewer.centralWidget()
        container_widget.addWidget(plot_widget)
        container_widget.addWidget(self._profile_tools)
        viewer.setCentralWidget(container_widget)
        self._profile_tools.enable()
        self._profile_tools.hide()

    def activate(self):
        if self._profile_tools.isVisible():
            self._profile_tools.hide()
        else:
            self._profile_tools.show()


class ProfileTools(QtWidgets.QWidget):

    fit_function = SelectionCallbackProperty()
    collapse_function = SelectionCallbackProperty()

    def __init__(self, parent=None):

        super(ProfileTools, self).__init__(parent=parent)

        self.ui = load_ui('profile_tools.ui', self,
                          directory=os.path.dirname(__file__))

        fix_tab_widget_fontsize(self.ui.tabs)

        self._viewer = weakref.ref(parent)
        self.image_viewer = None

    @property
    def viewer(self):
        return self._viewer()

    def show(self, *args):
        super(ProfileTools, self).show(*args)
        self._on_tab_change()

    def hide(self, *args):
        super(ProfileTools, self).hide(*args)
        self.rng_mode.deactivate()
        self.nav_mode.deactivate()

    def enable(self):

        self.nav_mode = NavigateMouseMode(self.viewer,
                                          press_callback=self._on_nav_activate)
        self.rng_mode = RangeMouseMode(self.viewer)

        self.nav_mode.state.add_callback('x', self._on_slider_change)

        self.ui.tabs.setCurrentIndex(0)

        self.ui.tabs.currentChanged.connect(self._on_tab_change)

        self.ui.button_settings.clicked.connect(self._on_settings)
        self.ui.button_fit.clicked.connect(self._on_fit)
        self.ui.button_clear.clicked.connect(self._on_clear)
        self.ui.button_collapse.clicked.connect(self._on_collapse)

        font = QtGui.QFont("Courier")
        font.setStyleHint(font.Monospace)
        self.ui.text_log.document().setDefaultFont(font)
        self.ui.text_log.setLineWrapMode(self.ui.text_log.NoWrap)

        self.axes = self.viewer.axes
        self.canvas = self.axes.figure.canvas

        self._fit_artists = []

        ProfileTools.fit_function.set_choices(self, list(fit_plugin))
        ProfileTools.fit_function.set_display_func(self, lambda fitter: fitter.label)
        self._connection_fit = connect_combo_selection(self, 'fit_function', self.ui.combosel_fit_function)

        ProfileTools.collapse_function.set_choices(self, list(COLLAPSE_FUNCS))
        ProfileTools.collapse_function.set_display_func(self, COLLAPSE_FUNCS.get)
        self._connection_collapse = connect_combo_selection(self, 'collapse_function', self.ui.combosel_collapse_function)

        self._toolbar_connected = False

        self.viewer.toolbar_added.connect(self._on_toolbar_added)

        self.viewer.state.add_callback('x_att', self._on_x_att_change)

    def _on_x_att_change(self, *event):
        self.nav_mode.clear()
        self.rng_mode.clear()

    def _on_nav_activate(self, *args):
        self._nav_data = self._visible_data()
        self._nav_viewers = {}
        for data in self._nav_data:
            pix_cid = is_convertible_to_single_pixel_cid(data, self.viewer.state.x_att_pixel)
            self._nav_viewers[data] = self._viewers_with_data_slice(data, pix_cid)

    def _on_slider_change(self, *args):

        x = self.nav_mode.state.x

        if x is None:
            return

        for data in self._nav_data:

            axis, slc = self._get_axis_and_pixel_slice(data, x)

            for viewer in self._nav_viewers[data]:
                slices = list(viewer.state.slices)
                slices[axis] = slc
                viewer.state.slices = tuple(slices)

    def _get_axis_and_pixel_slice(self, data, x):

        if self.viewer.state.x_att in data.pixel_component_ids:
            axis = self.viewer.state.x_att.axis
            slc = int(round(x))
        else:
            pix_cid = is_convertible_to_single_pixel_cid(data, self.viewer.state.x_att_pixel)
            axis = pix_cid.axis
            axis_view = [0] * data.ndim
            axis_view[pix_cid.axis] = slice(None)
            axis_values = data[self.viewer.state.x_att, axis_view]
            slc = int(np.argmin(np.abs(axis_values - x)))

        return axis, slc

    def _on_settings(self):
        d = FitSettingsWidget(self.fit_function())
        d.exec_()

    def _on_fit(self):
        """
        Fit a model to the data

        The fitting happens on a dedicated thread, to keep the UI
        responsive
        """

        if self.rng_mode.state.x_min is None or self.rng_mode.state.x_max is None:
            return

        x_range = self.rng_mode.state.x_range
        fitter = self.fit_function()

        def on_success(result):
            fit_results, x, y = result
            report = ""
            normalize = {}
            for layer_artist in fit_results:
                report += ("<b><font color='{0}'>{1}</font>"
                           "</b>".format(color2hex(layer_artist.state.color),
                                         layer_artist.layer.label))
                report += "<pre>" + fitter.summarize(fit_results[layer_artist], x, y) + "</pre>"
                if self.viewer.state.normalize:
                    normalize[layer_artist] = layer_artist.state.normalize_values
            self._report_fit(report)
            self._plot_fit(fitter, fit_results, x, y, normalize)

        def on_fail(exc_info):
            exc = '\n'.join(traceback.format_exception(*exc_info))
            self._report_fit("Error during fitting:\n%s" % exc)

        def on_done():
            self.ui.button_fit.setText("Fit")
            self.ui.button_fit.setEnabled(True)
            self.canvas.draw_idle()

        self.ui.button_fit.setText("Running...")
        self.ui.button_fit.setEnabled(False)

        w = Worker(self._fit, fitter, xlim=x_range)
        w.result.connect(on_success)
        w.error.connect(on_fail)
        w.finished.connect(on_done)

        self._fit_worker = w  # hold onto a reference
        w.start()

    def wait_for_fit(self):
        self._fit_worker.wait()

    def _report_fit(self, report):
        self.ui.text_log.document().setHtml(report)

    def _on_clear(self):
        self.ui.text_log.document().setPlainText('')
        self._clear_fit()
        self.canvas.draw_idle()

    def _fit(self, fitter, xlim=None):

        # We cycle through all the visible layers and get the plotted data
        # for each one of them.

        results = {}
        for layer in self.viewer.layers:
            if layer.enabled and layer.visible:
                x, y = layer.state.profile
                x = np.asarray(x)
                y = np.asarray(y)
                keep = (x >= min(xlim)) & (x <= max(xlim))
                if len(x) > 0:
                    results[layer] = fitter.build_and_fit(x[keep], y[keep])

        return results, x, y

    def _clear_fit(self):
        for artist in self._fit_artists[:]:
            artist.remove()
            self._fit_artists.remove(artist)

    def _plot_fit(self, fitter, fit_result, x, y, normalize):

        self._clear_fit()

        for layer in fit_result:
            # y_model = fitter.predict(fit_result[layer], x)
            self._fit_artists.append(fitter.plot(fit_result[layer], self.axes, x,
                                                 alpha=layer.state.alpha,
                                                 linewidth=layer.state.linewidth * 0.5,
                                                 color=layer.state.color,
                                                 normalize=normalize.get(layer, None))[0])

        self.canvas.draw_idle()

    def _visible_data(self):
        datasets = set()
        for layer_artist in self.viewer.layers:
            if layer_artist.enabled and layer_artist.visible:
                if isinstance(layer_artist.state.layer, BaseData):
                    datasets.add(layer_artist.state.layer)
                elif isinstance(layer_artist.state.layer, Subset):
                    datasets.add(layer_artist.state.layer.data)
        return list(datasets)

    def _viewers_with_data_slice(self, data, xatt):

        if self.viewer.session.application is None:
            return []

        viewers = []
        for tab in self.viewer.session.application.viewers:
            for viewer in tab:
                if isinstance(viewer, ImageViewer):
                    for layer_artist in viewer._layer_artist_container[data]:
                        if layer_artist.enabled and layer_artist.visible:
                            if len(viewer.state.slices) >= xatt.axis:
                                viewers.append(viewer)
        return viewers

    def _on_collapse(self):

        if self.rng_mode.state.x_min is None or self.rng_mode.state.x_max is None:
            return

        func = self.collapse_function
        x_range = self.rng_mode.state.x_range

        for data in self._visible_data():

            pix_cid = is_convertible_to_single_pixel_cid(data, self.viewer.state.x_att_pixel)

            for viewer in self._viewers_with_data_slice(data, pix_cid):

                slices = list(viewer.state.slices)

                # TODO: don't need to fetch axis twice
                axis, imin = self._get_axis_and_pixel_slice(data, x_range[0])
                axis, imax = self._get_axis_and_pixel_slice(data, x_range[1])

                current_slice = slices[axis]

                if isinstance(current_slice, AggregateSlice):
                    current_slice = current_slice.center

                imin, imax = min(imin, imax), max(imin, imax)

                slices[axis] = AggregateSlice(slice(imin, imax),
                                              current_slice,
                                              func)

                viewer.state.slices = tuple(slices)

    @property
    def mode(self):
        return MODES[self.tabs.currentIndex()]

    def _on_toolbar_added(self, *event):
        self.viewer.toolbar.tool_activated.connect(self._on_toolbar_activate)
        self.viewer.toolbar.tool_deactivated.connect(self._on_tab_change)

    def _on_toolbar_activate(self, *event):
        self.rng_mode.deactivate()
        self.nav_mode.deactivate()

    def _on_tab_change(self, *event):
        mode = self.mode
        if mode == 'navigate':
            self.rng_mode.deactivate()
            self.nav_mode.activate()
        else:
            self.rng_mode.activate()
            self.nav_mode.deactivate()



from __future__ import absolute_import, division, print_function

from mock import patch

from numpy.testing import assert_equal

from glue.core import Data, DataCollection, HubListener, ComponentID
from glue.core import message as msg
from glue.core.component_link import ComponentLink
from glue.core.parse import ParsedCommand, ParsedComponentLink

from ..component_arithmetic import ArithmeticEditorWidget, EquationEditorDialog


def auto_accept(equation):
    def exec_replacement(self):
        self.ui.expression.clear()
        self.ui.expression.insertPlainText(equation)
        self.accept()
    return exec_replacement


def auto_reject():
    def exec_replacement(self):
        self.ui.expression.clear()
        self.reject()
    return exec_replacement


class ChangeListener(HubListener):

    def __init__(self, data, *args, **kwargs):

        super(ChangeListener, self).__init__(*args, **kwargs)

        self.data = data
        self.removed = []
        self.added = []
        self.renamed = []
        self.reordered = []
        self.numeric = False

        self.register_to_hub(data.hub)

    def _has_data(self, message):
        return message.sender is self.data

    def register_to_hub(self, hub):

        hub.subscribe(self, msg.DataAddComponentMessage,
                      handler=self._component_added,
                      filter=self._has_data)

        hub.subscribe(self, msg.DataRemoveComponentMessage,
                      handler=self._component_removed,
                      filter=self._has_data)

        hub.subscribe(self, msg.DataRenameComponentMessage,
                      handler=self._component_renamed,
                      filter=self._has_data)

        hub.subscribe(self, msg.DataReorderComponentMessage,
                      handler=self._components_reordered,
                      filter=self._has_data)

        hub.subscribe(self, msg.NumericalDataChangedMessage,
                      handler=self._numerical_changed,
                      filter=self._has_data)

    def _component_added(self, message):
        self.added.append(message.component_id)

    def _component_removed(self, message):
        self.removed.append(message.component_id)

    def _component_renamed(self, message):
        self.renamed.append(message.component_id)

    def _components_reordered(self, message):
        self.reordered = message.component_ids

    def _numerical_changed(self, message):
        self.numeric = True

    def assert_exact_changes(self, added=[], removed=[], renamed=[], reordered=[], numerical=False):
        assert set(added) == set(self.added)
        assert set(removed) == set(self.removed)
        assert set(renamed) == set(self.renamed)
        assert reordered == self.reordered
        assert numerical is self.numeric


class TestArithmeticEditorWidget:

    def setup_method(self):

        self.data1 = Data(x=[1, 2, 3], y=[3.5, 4.5, -1.0], z=['a', 'r', 'w'])
        self.data2 = Data(a=[3, 4, 1], b=[1.5, -2.0, 3.5], c=['y', 'e', 'r'])

        # Add a derived component so that we can test how we deal with existing ones
        components = dict((cid.label, cid) for cid in self.data2.components)
        pc = ParsedCommand('{a}', components)
        link = ParsedComponentLink(ComponentID('d'), pc)
        self.data2.add_component_link(link)

        self.data_collection = DataCollection([self.data1, self.data2])

        link = ComponentLink([self.data1.id['x']], self.data2.id['a'])
        self.data_collection.add_link(link)

        self.listener1 = ChangeListener(self.data1)
        self.listener2 = ChangeListener(self.data2)

    def test_nochanges(self):
        editor = ArithmeticEditorWidget(self.data_collection)
        editor.show()
        editor.button_ok.click()
        self.listener1.assert_exact_changes()
        self.listener2.assert_exact_changes()
        editor.close()

    def test_add_derived_and_rename(self):
        editor = ArithmeticEditorWidget(self.data_collection)
        editor.show()
        with patch.object(EquationEditorDialog, 'exec_', auto_accept('{x} + {y}')):
            editor.button_add_derived.click()
        item = list(editor.list)[0]
        item.setText(0, 'new')
        editor.button_ok.click()
        self.listener1.assert_exact_changes(added=[self.data1.id['new']])
        self.listener2.assert_exact_changes()
        assert_equal(self.data1['new'], [4.5, 6.5, 2.0])
        editor.close()

    def test_add_derived_and_cancel(self):
        editor = ArithmeticEditorWidget(self.data_collection)
        editor.show()
        with patch.object(EquationEditorDialog, 'exec_', auto_reject()):
            editor.button_add_derived.click()
        assert len(editor.list) == 0
        editor.close()

    def test_edit_existing_equation(self):
        assert_equal(self.data2['d'], [3, 4, 1])
        editor = ArithmeticEditorWidget(self.data_collection)
        editor.show()
        assert len(editor.list) == 0
        editor.combosel_data.setCurrentIndex(1)
        assert len(editor.list) == 1
        editor.list.select_cid(self.data2.id['d'])
        with patch.object(EquationEditorDialog, 'exec_', auto_accept('{a} + {b}')):
            editor.button_edit_derived.click()
        editor.button_ok.click()
        self.listener1.assert_exact_changes()
        self.listener2.assert_exact_changes(numerical=True)
        assert_equal(self.data2['d'], [4.5, 2.0, 4.5])
        editor.close()

    # TODO: add an updated version of the following test back once we add
    # support for using derived components in derived components.
    #
    # def test_edit_equation_after_rename(self):
    #     editor = ArithmeticEditorWidget(self.data_collection)
    #     editor.show()
    #     editor.combosel_data.setCurrentIndex(1)
    #     editor.list['main'].select_cid(self.data2.id['a'])
    #     editor.list['main'].selected_item.setText(0, 'renamed')
    #     editor.list.select_cid(self.data2.id['d'])
    #     with patch.object(EquationEditorDialog, 'exec_', auto_accept('{renamed} + 1')):
    #         editor.button_edit_derived.click()
    #     editor.button_ok.click()
    #     self.listener1.assert_exact_changes()
    #     self.listener2.assert_exact_changes(renamed=[self.data2.id['renamed']], numerical=True)
    #     assert_equal(self.data2['d'], [4, 5, 2])



# pylint: disable=I0011,W0613,W0201,W0212,E1101,E1103

from __future__ import absolute_import, division, print_function

import pytest
import numpy as np
from mock import MagicMock
from numpy.testing import assert_array_equal, assert_equal

from ..coordinates import Coordinates
from ..component_link import ComponentLink
from ..link_helpers import LinkSame
from ..data import Data, Component, ComponentID, DerivedComponent
from ..data_collection import DataCollection
from ..hub import HubListener
from ..message import (Message, DataCollectionAddMessage, DataRemoveComponentMessage,
                       DataCollectionDeleteMessage, DataAddComponentMessage,
                       ComponentsChangedMessage, PixelAlignedDataChangedMessage)
from ..exceptions import IncompatibleAttribute

from .test_state import clone


class HubLog(HubListener):

    def __init__(self):
        self.messages = []

    def register_to_hub(self, hub):
        hub.subscribe(self, Message)

    def notify(self, message):
        self.messages.append(message)

    def clear(self):
        self.messages[:] = []

    def messages_by_type(self, klass):
        return [msg for msg in self.messages if isinstance(msg, klass)]


class TestDataCollection(object):

    def setup_method(self, method):
        self.dc = DataCollection()
        self.data = MagicMock(spec_set=Data)
        self.hub = self.dc.hub
        self.log = HubLog()
        self.log.register_to_hub(self.hub)

    def test_init_scalar(self):
        """Single data object passed to init adds to collection"""
        d = Data()
        dc = DataCollection(d)
        assert d in dc

    def test_init_list(self):
        """List of data objects passed to init auto-added to collection"""
        d1 = Data()
        d2 = Data()
        dc = DataCollection([d1, d2])
        assert d1 in dc
        assert d2 in dc

    def test_data(self):
        """ data attribute is a list of all appended data"""
        self.dc.append(self.data)
        assert self.dc.data == [self.data]

    def test_append(self):
        """ append method adds to collection """
        self.dc.append(self.data)
        assert self.data in self.dc

    def test_multi_append(self):
        """ append method works with lists """
        d = Data('test1', x=[1, 2, 3])
        d2 = Data('test2', y=[2, 3, 4])
        self.dc.append([d, d2])
        assert d in self.dc
        assert d2 in self.dc

    def test_ignore_multi_add(self):
        """ data only added once, even after multiple calls to append """
        self.dc.append(self.data)
        self.dc.append(self.data)
        assert len(self.dc) == 1

    def test_remove(self):
        self.dc.append(self.data)
        self.dc.remove(self.data)
        assert self.data not in self.dc

    def test_ignore_multi_remove(self):
        self.dc.append(self.data)
        self.dc.remove(self.data)
        self.dc.remove(self.data)
        assert self.data not in self.dc

    def test_append_broadcast(self):
        """ Call to append generates a DataCollectionAddMessage """
        self.dc.append(self.data)
        msg = self.log.messages[-1]
        assert msg.sender == self.dc
        assert isinstance(msg, DataCollectionAddMessage)
        assert msg.data is self.data

    def test_remove_broadcast(self):
        """ call to remove generates a DataCollectionDeleteMessage """
        self.dc.append(self.data)
        self.dc.remove(self.data)
        msg = self.log.messages[-1]
        assert msg.sender == self.dc
        assert isinstance(msg, DataCollectionDeleteMessage)
        assert msg.data is self.data

    def test_register_assigns_hub_of_data(self):
        self.dc.append(self.data)
        self.data.register_to_hub.assert_called_once_with(self.hub)

    def test_get_item(self):
        self.dc.append(self.data)
        assert self.dc[0] is self.data

    def test_iter(self):
        self.dc.append(self.data)
        assert set(self.dc) == set([self.data])

    def test_len(self):
        assert len(self.dc) == 0
        self.dc.append(self.data)
        assert len(self.dc) == 1
        self.dc.append(self.data)
        assert len(self.dc) == 1
        self.dc.remove(self.data)
        assert len(self.dc) == 0

    def test_derived_links_autoadd(self):
        """
        When appending a data set, its DerivedComponents should be ingested into
        the LinkManager
        """
        d = Data()
        id1 = ComponentID("id1")
        id2 = ComponentID("id2")
        link = ComponentLink([id1], id2)
        dc = DerivedComponent(d, link)
        d.add_component(Component(np.array([1, 2, 3])), id1)
        d.add_component(dc, id2)

        dc = DataCollection()
        dc.append(d)

        assert link in dc._link_manager

    def test_catch_data_add_component_message(self):
        """
        DerviedAttributes added to a dataset in a collection
        should generate messages that the collection catches.
        """
        d = Data()
        id1 = ComponentID("id1")
        id2 = ComponentID("id2")
        link = ComponentLink([id1], id2)
        dc = DerivedComponent(d, link)

        self.dc.append(d)
        d.add_component(Component(np.array([1, 2, 3])), id1)
        assert link not in self.dc._link_manager
        self.log.clear()
        d.add_component(dc, id2)

        assert link in self.dc._link_manager

        msgs = sorted(self.log.messages, key=lambda x: str(type(x)))

        assert isinstance(msgs[0], ComponentsChangedMessage)
        assert isinstance(msgs[1], DataAddComponentMessage)

    def test_links_auto_added(self):
        id1 = ComponentID("id1")
        id2 = ComponentID("id2")
        link = ComponentLink([id1], id2)
        self.data.links = [link]
        self.dc.append(self.data)
        assert link in self.dc.links

    def test_add_links(self):
        """ links attribute behaves like an editable list """
        id1 = ComponentID("id1")
        id2 = ComponentID("id2")
        link = ComponentLink([id1], id2)
        self.dc.set_links([link])
        assert link in self.dc.links

    def test_add_links_updates_components(self):
        """
        Setting links attribute automatically adds components to data
        """
        d = Data()
        comp = Component(np.array([1, 2, 3]))
        id1 = ComponentID("id1")
        d.add_component(comp, id1)
        id2 = ComponentID("id2")
        self.dc.append(d)
        link = ComponentLink([id1], id2)
        self.dc.set_links([link])
        assert_equal(d[id2], d[id1])

    def test_links_propagated(self):
        """Web of links is grown and applied to data automatically"""
        from ..component_link import ComponentLink
        d = Data()
        dc = DataCollection([d])

        cid1 = d.add_component(np.array([1, 2, 3]), 'a')
        cid2 = ComponentID('b')
        cid3 = ComponentID('c')

        links1 = ComponentLink([cid1], cid2)
        dc.add_link(links1)

        assert_equal(d[cid2], d[cid1])

        links2 = ComponentLink([cid2], cid3)
        dc.add_link(links2)
        assert_equal(d[cid3], d[cid2])

        dc.remove_link(links2)
        with pytest.raises(IncompatibleAttribute):
            d[cid3]
        assert_equal(d[cid2], d[cid1])

        dc.remove_link(links1)
        with pytest.raises(IncompatibleAttribute):
            d[cid2]

    def test_merge_links(self):
        """Trivial links should be merged, discarding the duplicate ID"""
        d1 = Data(x=[1, 2, 3])
        d2 = Data(x=[2, 3, 4])
        dc = DataCollection([d1, d2])

        original_id = d2.id['x']
        link = ComponentLink([d1.id['x']], d2.id['x'])
        dc.add_link(link)

        # NOTE: the behavior tested here is not desirable anymore, so the relevant
        #       parts that are no longer true have been commented out and replaced
        #       by the new behavior.

        # assert d1.id['x'] is not original_id
        # assert duplicated_id not in d2.components
        assert d1.id['x'] is not original_id
        assert d2.id['x'] is original_id
        assert original_id in d2.components

        assert_array_equal(d1[d1.id['x']], [1, 2, 3])
        assert_array_equal(d2[d1.id['x']], [2, 3, 4])

    def test_merge(self):

        x = Data(x=[1, 2, 3])
        y = Data(y=[2, 3, 4])
        dc = DataCollection([x, y])

        dc.merge(x, y)

        assert x not in dc
        assert y not in dc

        assert_array_equal(dc[0]['x'], [1, 2, 3])
        assert_array_equal(dc[0]['y'], [2, 3, 4])

    def test_merge_discards_duplicate_pixel_components(self):
        x = Data(x=[1, 2, 3])
        y = Data(y=[2, 3, 4])
        dc = DataCollection([x, y])
        dc.merge(x, y)
        merged = dc[0]

        assert len(merged.coordinate_components) == 2
        assert y.pixel_component_ids[0] not in merged.components

    def test_merge_forbids_single_argument(self):
        x = Data(x=[1, 2, 3])
        y = Data(y=[2, 3, 4])
        dc = DataCollection([x, y])
        with pytest.raises(ValueError) as exc:
            dc.merge(x)
        assert exc.value.args[0] == 'merge requires 2 or more arguments'

    def test_merge_requires_same_shapes(self):
        x = Data(x=[1, 2, 3])
        y = Data(y=[2, 3, 4, 5])
        dc = DataCollection([x, y])
        with pytest.raises(ValueError) as exc:
            dc.merge(x, y)
        assert exc.value.args[0] == 'All arguments must have the same shape'

    def test_merge_disambiguates_components(self):
        x = Data(x=[1, 2, 3])
        old = set(x.components)
        y = Data(x=[2, 3, 4])
        dc = DataCollection([x, y])
        dc.merge(x, y)

        z = dc[0]
        new = list(set(z.components) - old)[0]

        assert new.label != 'x'

    def test_merge_multiargument(self):

        dc = DataCollection([Data(x=[1, 2, 3]),
                             Data(y=[2, 3, 4]),
                             Data(z=[3, 4, 5])])

        dc.merge(*list(dc))
        assert len(dc) == 1
        d = dc[0]
        assert_array_equal(d['y'], [2, 3, 4])
        assert_array_equal(d['z'], [3, 4, 5])

    def test_merging_preserves_links_forwards(self):

        a = Data(a=[1, 2, 3])
        b = Data(b=[2, 3, 4])
        c = Data(c=[3, 4, 5])

        dc = DataCollection([a, b, c])
        dc.add_link(ComponentLink([a.id['a']], b.id['b'], lambda x: x))
        dc.add_link(ComponentLink([b.id['b']], c.id['c'], lambda x: x))

        assert_array_equal(a['c'], [1, 2, 3])
        dc.merge(a, b)
        assert_array_equal(a['c'], [1, 2, 3])

    def test_merging_preserves_links_backwards(self):

        a = Data(a=[1, 2, 3], label='Data A')
        b = Data(b=[2, 3, 4], label='Data B')
        c = Data(c=[3, 4, 5], label='Data C')

        dc = DataCollection([a, b, c])
        dc.add_link(ComponentLink([c.id['c']], b.id['b'], lambda x: x))
        dc.add_link(ComponentLink([b.id['b']], a.id['a'], lambda x: x))

        assert_array_equal(c['a'], [3, 4, 5])
        dc.merge(a, b)
        assert_array_equal(c['a'], [3, 4, 5])

    def test_merge_coordinates(self):

        # Regression test to make sure the coordinates from the first dataset
        # are preserved.

        x = Data(x=[1, 2, 3])
        y = Data(y=[2, 3, 4])
        dc = DataCollection([x, y])

        x.coords = Coordinates()
        y.coords = Coordinates()

        dc.merge(x, y)

        assert dc[0].coords is x.coords

    def test_merge_coordinates_preserve_labels(self):

        # Regression test to make sure that axis labels are preserved after
        # merging.

        x = Data(x=[1, 2, 3])
        y = Data(y=[2, 3, 4])
        dc = DataCollection([x, y])

        class CustomCoordinates(Coordinates):
            def axis_label(self, axis):
                return 'Custom {0}'.format(axis)

        x.coords = CustomCoordinates()
        y.coords = CustomCoordinates()

        dc.merge(x, y)

        assert sorted(cid.label for cid in dc[0].world_component_ids) == ['Custom 0']

    def test_merge_visible_components(self):
        # Regression test for a bug that caused visible_components to be empty
        # for a dataset made from merging other datasets.
        x = Data(x=[1, 2, 3], label='dx')
        y = Data(y=[2, 3, 4], label='dy')
        dc = DataCollection([x, y])
        dc.merge(x, y)
        assert dc[0].main_components[0] is x.id['x']
        assert dc[0].main_components[1] is y.id['y']

    def test_remove_component_message(self):

        # Regression test to make sure that removing a component emits the
        # appropriate messages.

        data = Data(x=[1, 2, 3], y=[4, 5, 6])
        self.dc.append(data)

        remove_id = data.id['y']

        self.log.clear()

        data.remove_component(remove_id)

        msgs = sorted(self.log.messages, key=lambda x: str(type(x)))

        print([type(msg) for msg in msgs])

        assert isinstance(msgs[0], ComponentsChangedMessage)

        assert isinstance(msgs[1], DataRemoveComponentMessage)
        assert msgs[1].component_id is remove_id

    def test_links_preserved_session(self):

        # This tests that the separation of internal vs external links is
        # preserved in session files.

        d1 = Data(a=[1, 2, 3])
        d2 = Data(b=[2, 3, 4])

        dc = DataCollection([d1, d2])
        dc.add_link(ComponentLink([d2.id['b']], d1.id['a']))

        d1['x'] = d1.id['a'] + 1

        assert len(d1.coordinate_links) == 2
        assert len(d1.derived_links) == 1
        assert len(dc._link_manager._external_links) == 1

        dc2 = clone(dc)

        assert len(dc2[0].coordinate_links) == 2
        assert len(dc2[0].derived_links) == 1
        assert len(dc2._link_manager._external_links) == 1

    def test_pixel_aligned(self):

        data1 = Data(x=np.ones((2, 4, 3)))
        data2 = Data(y=np.ones((4, 3, 2)))
        data3 = Data(z=np.ones((2, 3)))

        self.dc.extend([data1, data2, data3])

        # Add one set of links, which shouldn't change anything

        self.dc.add_link(LinkSame(data1.pixel_component_ids[0], data2.pixel_component_ids[2]))
        self.dc.add_link(LinkSame(data1.pixel_component_ids[0], data3.pixel_component_ids[0]))

        assert len(data1.pixel_aligned_data) == 0
        assert len(data2.pixel_aligned_data) == 0
        assert len(data3.pixel_aligned_data) == 0
        assert len(self.log.messages_by_type(PixelAlignedDataChangedMessage)) == 0

        # Add links between a second set of axes

        self.dc.add_link(LinkSame(data1.pixel_component_ids[2], data2.pixel_component_ids[1]))
        self.dc.add_link(LinkSame(data1.pixel_component_ids[2], data3.pixel_component_ids[1]))

        # At this point, data3 has all its axes contained in data1 and data2

        assert len(data1.pixel_aligned_data) == 0
        assert len(data2.pixel_aligned_data) == 0
        assert len(data3.pixel_aligned_data) == 2
        assert data3.pixel_aligned_data[data1] == [0, 2]
        assert data3.pixel_aligned_data[data2] == [2, 1]
        messages = self.log.messages_by_type(PixelAlignedDataChangedMessage)
        assert len(messages) == 1
        assert messages[0].data is data3

        # Finally we can link the third set of axes

        self.dc.add_link(LinkSame(data1.pixel_component_ids[1], data2.pixel_component_ids[0]))

        # At this point, data1 and data2 should now be linked

        assert len(data1.pixel_aligned_data) == 1
        assert data1.pixel_aligned_data[data2] == [2, 0, 1]
        assert len(data2.pixel_aligned_data) == 1
        assert data2.pixel_aligned_data[data1] == [1, 2, 0]
        assert len(data3.pixel_aligned_data) == 2
        assert data3.pixel_aligned_data[data1] == [0, 2]
        assert data3.pixel_aligned_data[data2] == [2, 1]
        messages = self.log.messages_by_type(PixelAlignedDataChangedMessage)
        assert len(messages) == 3
        assert messages[1].data is data1
        assert messages[2].data is data2

    def test_merge_duplicate(self):

        x = Data(x=[1, 2, 3], label='d1')
        y = Data(x=[2, 3, 4], label='d2')
        z = Data(y=[2, 3, 4], label='d3')
        dc = DataCollection([x, y, z])

        dc.merge(x, y, z)

        assert dc[0].main_components[0].label == 'x [d1]'
        assert dc[0].main_components[1].label == 'x [d2]'
        assert dc[0].main_components[2].label == 'y'

    def test_suggest_merge_label(self):

        x = Data(x=[1, 2, 3], label='abc_1')
        y = Data(x=[2, 3, 4], label='abc_2')
        z = Data(y=[2, 3, 4], label='abc_3')
        w = Data(y=[2, 3, 4], label='banana')
        dc = DataCollection([x, y, z])

        assert dc.suggest_merge_label(x, y) == 'abc'
        assert dc.suggest_merge_label(x, y, z) == 'abc'
        assert dc.suggest_merge_label(x, y, z, w) == 'Merged data'

        u = Data(y=[2, 3, 4], label='abc')
        dc.append(u)

        assert dc.suggest_merge_label(x, y) == 'abc [2]'

        v = Data(y=[2, 3, 4], label='Merged data')
        dc.append(v)

        assert dc.suggest_merge_label(x, w) == 'Merged data [2]'



from __future__ import absolute_import, division, print_function

import numpy as np

from matplotlib.patches import Rectangle

from glue.viewers.matplotlib.mpl_axes import update_appearance_from_settings
from glue.external.echo import delay_callback
from glue.utils import mpl_to_datetime64

__all__ = ['MatplotlibViewerMixin']

SCRIPT_HEADER = """
# Initialize figure
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1, aspect='{aspect}')
""".strip()

SCRIPT_FOOTER = """
# Set limits
ax.set_xlim({x_min}, {x_max})
ax.set_ylim({y_min}, {y_max})

# Set scale (log or linear)
ax.set_xscale('{x_log_str}')
ax.set_yscale('{y_log_str}')

# Set axis label properties
ax.set_xlabel('{x_axislabel}', weight='{x_axislabel_weight}', size={x_axislabel_size})
ax.set_ylabel('{y_axislabel}', weight='{y_axislabel_weight}', size={y_axislabel_size})

# Set tick label properties
ax.tick_params('x', labelsize={x_ticklabel_size})
ax.tick_params('y', labelsize={x_ticklabel_size})

# Save figure
fig.savefig('glue_plot.png')
plt.close(fig)
""".strip()

ZORDER_MAX = 100000


class MatplotlibViewerMixin(object):

    def setup_callbacks(self):

        for spine in self.axes.spines.values():
            spine.set_zorder(ZORDER_MAX)

        self.loading_rectangle = Rectangle((0, 0), 1, 1, color='0.9', alpha=0.9,
                                           zorder=ZORDER_MAX - 1, transform=self.axes.transAxes)
        self.loading_rectangle.set_visible(False)
        self.axes.add_patch(self.loading_rectangle)

        self.loading_text = self.axes.text(0.4, 0.5, 'Computing', color='k',
                                           zorder=self.loading_rectangle.get_zorder() + 1,
                                           ha='left', va='center',
                                           transform=self.axes.transAxes)
        self.loading_text.set_visible(False)

        self.state.add_callback('x_min', self.limits_to_mpl)
        self.state.add_callback('x_max', self.limits_to_mpl)
        self.state.add_callback('y_min', self.limits_to_mpl)
        self.state.add_callback('y_max', self.limits_to_mpl)

        if (self.state.x_min or self.state.x_max or self.state.y_min or self.state.y_max) is None:
            self.limits_from_mpl()
        else:
            self.limits_to_mpl()

        self.state.add_callback('x_log', self.update_x_log, priority=1000)
        self.state.add_callback('y_log', self.update_y_log, priority=1000)

        self.update_x_log()

        self.axes.callbacks.connect('xlim_changed', self.limits_from_mpl)
        self.axes.callbacks.connect('ylim_changed', self.limits_from_mpl)
        self.figure.canvas.mpl_connect('resize_event', self._on_resize)

        self._on_resize()

        self.axes.set_autoscale_on(False)

        self.state.add_callback('x_axislabel', self.update_x_axislabel)
        self.state.add_callback('x_axislabel_weight', self.update_x_axislabel)
        self.state.add_callback('x_axislabel_size', self.update_x_axislabel)

        self.state.add_callback('y_axislabel', self.update_y_axislabel)
        self.state.add_callback('y_axislabel_weight', self.update_y_axislabel)
        self.state.add_callback('y_axislabel_size', self.update_y_axislabel)

        self.state.add_callback('x_ticklabel_size', self.update_x_ticklabel)
        self.state.add_callback('y_ticklabel_size', self.update_y_ticklabel)

        self.update_x_axislabel()
        self.update_y_axislabel()
        self.update_x_ticklabel()
        self.update_y_ticklabel()

    def _update_computation(self, message=None):
        pass

    def update_x_axislabel(self, *event):
        self.axes.set_xlabel(self.state.x_axislabel,
                             weight=self.state.x_axislabel_weight,
                             size=self.state.x_axislabel_size)
        self.redraw()

    def update_y_axislabel(self, *event):
        self.axes.set_ylabel(self.state.y_axislabel,
                             weight=self.state.y_axislabel_weight,
                             size=self.state.y_axislabel_size)
        self.redraw()

    def update_x_ticklabel(self, *event):
        self.axes.tick_params(axis='x', labelsize=self.state.x_ticklabel_size)
        self.axes.xaxis.get_offset_text().set_fontsize(self.state.x_ticklabel_size)
        self.redraw()

    def update_y_ticklabel(self, *event):
        self.axes.tick_params(axis='y', labelsize=self.state.y_ticklabel_size)
        self.axes.yaxis.get_offset_text().set_fontsize(self.state.y_ticklabel_size)
        self.redraw()

    def redraw(self):
        self.figure.canvas.draw_idle()

    def update_x_log(self, *args):
        self.axes.set_xscale('log' if self.state.x_log else 'linear')
        self.redraw()

    def update_y_log(self, *args):
        self.axes.set_yscale('log' if self.state.y_log else 'linear')
        self.redraw()

    def limits_from_mpl(self, *args, **kwargs):

        if getattr(self, '_skip_limits_from_mpl', False):
            return

        if isinstance(self.state.x_min, np.datetime64):
            x_min, x_max = [mpl_to_datetime64(x) for x in self.axes.get_xlim()]
        else:
            x_min, x_max = self.axes.get_xlim()

        if isinstance(self.state.y_min, np.datetime64):
            y_min, y_max = [mpl_to_datetime64(y) for y in self.axes.get_ylim()]
        else:
            y_min, y_max = self.axes.get_ylim()

        with delay_callback(self.state, 'x_min', 'x_max', 'y_min', 'y_max'):

            self.state.x_min = x_min
            self.state.x_max = x_max
            self.state.y_min = y_min
            self.state.y_max = y_max

    def limits_to_mpl(self, *args):

        if (self.state.x_min is None or self.state.x_max is None or
                self.state.y_min is None or self.state.y_max is None):
            return

        x_min, x_max = self.state.x_min, self.state.x_max

        if self.state.x_log:
            if self.state.x_max <= 0:
                x_min, x_max = 0.1, 1
            elif self.state.x_min <= 0:
                x_min = x_max / 10

        y_min, y_max = self.state.y_min, self.state.y_max
        if self.state.y_log:
            if self.state.y_max <= 0:
                y_min, y_max = 0.1, 1
            elif self.state.y_min <= 0:
                y_min = y_max / 10

        # Since we deal with aspect ratio internally, there are no conditions
        # under which we would want to immediately change the state once the
        # limits are set in Matplotlib, so we avoid this by setting the
        # _skip_limits_from_mpl attribute which is then recognized inside
        # limits_from_mpl
        self._skip_limits_from_mpl = True
        try:
            self.axes.set_xlim(x_min, x_max)
            self.axes.set_ylim(y_min, y_max)
            self.axes.figure.canvas.draw_idle()
        finally:
            self._skip_limits_from_mpl = False

    @property
    def axes_ratio(self):

        # Get figure aspect ratio
        width, height = self.figure.get_size_inches()
        fig_aspect = height / width

        # Get axes aspect ratio
        l, b, w, h = self.axes.get_position().bounds
        axes_ratio = fig_aspect * (h / w)

        return axes_ratio

    def _on_resize(self, *args):
        self.state._set_axes_aspect_ratio(self.axes_ratio)

    def _update_appearance_from_settings(self, message=None):
        update_appearance_from_settings(self.axes)
        self.redraw()

    def get_layer_artist(self, cls, layer=None, layer_state=None):
        return cls(self.axes, self.state, layer=layer, layer_state=layer_state)

    def apply_roi(self, roi, override_mode=None):
        """ This method must be implemented by subclasses """
        raise NotImplementedError

    def _script_header(self):
        state_dict = self.state.as_dict()
        return ['import matplotlib', "matplotlib.use('Agg')", 'import matplotlib.pyplot as plt'], SCRIPT_HEADER.format(**state_dict)

    def _script_footer(self):
        state_dict = self.state.as_dict()
        state_dict['x_log_str'] = 'log' if self.state.x_log else 'linear'
        state_dict['y_log_str'] = 'log' if self.state.y_log else 'linear'
        return [], SCRIPT_FOOTER.format(**state_dict)



import os
from .. import icon_path


def test_icon_path():

    path = icon_path('glue_replace')
    assert os.path.exists(path)

    path = icon_path('glue_replace', icon_format='png')
    assert os.path.exists(path)

    path = icon_path('glue_replace', icon_format='svg')
    assert os.path.exists(path)
